# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import json
import os, sys

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from pypapi import *
# get all dson class
from pypapi.dson_test import DsonTest
from pypapi.building.building import DsonBuilding


dson_path = os.path.join(os.getcwd(), "dson")

if __name__ == "__main__":
    index = 0
    if not os.path.exists(dson_path):
        os.makedirs(dson_path, exist_ok=True)
    for cls_name, cls in DsonManager().dsonMap.items():
        for name, value in vars(cls).items():
            if isinstance(value, nextpcgmethod):
                func = value.__func__
                dson_meta_info = DsonMetaInfo(cls_name)
                json_data = create_dson_from_pda(func, cls.label, dson_meta_info)
                file_name = name + '.dson'
                file_path = os.path.join(dson_path, file_name).replace('\\', '/')
                fp = open(file_path, "w")
                json.dump(json_data, fp)
                index += 1



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

from .meta_helper import Singleton
from .const import *


class nextpcgmethod(staticmethod):
    def __init__(self, function):
        super(nextpcgmethod, self).__init__(function)


class DsonMetaInfo:
    def __init__(self, cls_name):
        self.cls_name = cls_name

    def to_json(self):
        return {dson_meta_clsname_tag: self.cls_name}

    @staticmethod
    def from_json(json_data):
        return DsonMetaInfo(json_data[dson_meta_clsname_tag])


class DsonManager(metaclass=Singleton):
    def __init__(self):
        self.dsonMap = {}


class DsonMeta(type):
    def __init__(cls, *args, **kwargs):
        assert hasattr(cls, 'label')
        DsonManager().dsonMap[cls.__name__] = cls
        super().__init__(*args, **kwargs)


class DsonBase(metaclass=DsonMeta):
    label = "pda"

    def __init__(self):
        pass



# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

from .field import Field, FieldCategory, FieldDesc
from .geo_heightfield import HeightField
from typing import List


class HeightFieldField(Field):
    """
    _value is list of HeightField
    """
    field_type_name = "paramType.HeightField"
    support_param = False

    # @classmethod
    # def create_json_field(cls, field_category=FieldCategory.Input, name: str = None):
    #     """ create a json object
    #     """
    #     """get json format field"""
    #     field = {'Type': cls.field_type_name, 'If List': cls.if_list, 'FieldCategory': field_category.name}
    #     if name is not None:
    #         field["Label"] = name
    #     return field

    def __init__(self, value=None):
        self._value: List[HeightField] = value

    @classmethod
    def from_json(cls, json_object, input_name, field_desc):
        if field_desc.work_path is None:
            raise Exception("HeightFieldField from_json failed:work path need to be set, not allowed")
        value = []
        for editing_layer_key, editing_layer_val in json_object.items():
            if 'EditingLayer_' in editing_layer_key:
                heightfield = HeightField.create_from_json(json_object, editing_layer_val, field_desc.work_path, field_desc.io_mode, field_desc.raw_compress)
                value.append(heightfield)
        return cls(value)

    def get_input(self):
        return self._value

    def get_output(self, json_field, field_desc: FieldDesc, data_field='Value Data', output_index=0):
        geo_index = 0
        for heightfield in self._value:
            num = heightfield.to_json(geo_index, output_index, json_field, field_desc.io_mode, field_desc.work_path, field_desc.raw_compress)
            geo_index += num




# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import numpy as np

from .dson import DsonBase, nextpcgmethod
from .field import *
from .field_heightfield import HeightFieldField
from .field_instanced_staticmesh import InstancedStaticMeshField
from .field_instanced_staticmesh import InstancedStaticMesh


def sum_impl(eles: list, coff: int) -> int:
    return sum(eles) * coff


class DsonTest(DsonBase):
    
    @nextpcgmethod
    def test_many(a: Int4, string: String, b: Bool = False, c: Float3 = (1.1, 1.2, 1.3)) -> Tuple[Int, String, Float4]:
        a_v = a.get_input()
        str_v = string.get_input()
        b_v = b.get_input()
        c_v = c.get_input()
        re1 = a_v[0]
        re2 = str_v + ": from server"
        c_v.append(3.141)
        return Int(re1), String(re2), Float4(c_v)

    @nextpcgmethod
    def test_float3(a: Float3, b: Float3) -> Float3:
        a_in: List = a.get_input()
        b_in = b.get_input()
        c_in = a_in.copy()
        c_in[0] += b_in[0]
        c_in[1] += b_in[1]
        c_in[2] += b_in[2]

        return Float3(c_in)

    @nextpcgmethod
    def test_sum(elements: Int3, coff: Int) -> Tuple[Int, String]:
        result = sum_impl(elements.get_input(),
                          coff.get_input())
        words = "来自python的问候"
        return Int(result), String(words)

    @nextpcgmethod
    def test_add_int(a: Int, b: Int) -> Int:
        return Int(a.get_input() + b.get_input())

    @nextpcgmethod
    def test_file(file: FileField, a: Int) -> Int:
        return Int(3)

    @nextpcgmethod
    def test_file_echo(file: FileField) -> FileField:
        return file

    @nextpcgmethod
    def test_json(json1: JsonField) -> Tuple[Int, JsonField]:
        json_dict = json1.get_input()
        re1 = json_dict['a']
        re2 = {"a": 12}
        return Int(re1), JsonField(re2, "ff")

    @nextpcgmethod
    def test_float3list(in_floats: ListField[Float3]) -> ListField[Float3]:
        re_float3s = []
        for float3 in in_floats.get_input():
            float3[1] += 10
            print(float3)
            re_float3s.append(float3)
        return ListField[Float3](re_float3s)

    @nextpcgmethod
    def test_params(a: Int, b: Int = 1) -> Int:
        a_v = a.get_input()
        b_v = b.get_input()
        c_v = a_v + b_v
        return Int(c_v)

    @nextpcgmethod
    def test_float(a: Float, b: Float4) -> Float2:
        a_v = a.get_input()
        b_v = b.get_input()
        c_v = (a_v + b_v[0], b_v[1] + b_v[2] + b_v[3])
        return Float2(c_v)

    @nextpcgmethod
    def test_float1(a: Float) -> Float:
        return a

    @nextpcgmethod
    def test_int2(a: Int3 = (1, 2, 3)) -> Int3:
        return a

    @nextpcgmethod
    def test_bool(a: Bool = True) -> Bool:
        return Bool(not a.get_input())

    @nextpcgmethod
    def test_string(string: String) -> String:
        str_v = string.get_input()
        str_v += "return from server"
        return String(str_v)

    @nextpcgmethod
    def test_list_int(aa: ListField[Int]) -> ListField[Int]:
        return aa

    @nextpcgmethod
    def test_float2_list(aa: ListField[Float2]) -> ListField[Float2]:
        aa_inner = aa.get_input()
        aa_inner.append([3.45, 2.3])
        return ListField[Float2](aa_inner)

    @nextpcgmethod
    def test_string(in_str: String) -> Tuple[String, ListField[String]]:
        a = in_str.get_input()
        out_str = a + '_ hi from server'
        return String(a), ListField([a, out_str], String)

    @nextpcgmethod
    def test_jsonfiles(in_files: ListField[JsonField]) -> ListField[JsonField]:
        re1 = {"a": 2}
        re2 = {"b": True}
        result = ListField[JsonField]([re1, re2])
        return result

    @nextpcgmethod
    def test_heightfield(in_hf: HeightFieldField) -> HeightFieldField:
        a = in_hf.get_input()
        hf = a[0].volumes['layer0-height'].mat
        hf[2][10][0] += 20
        return HeightFieldField(a)

    @nextpcgmethod
    def test_instanced_staticmesh(in_ism: InstancedStaticMeshField) -> InstancedStaticMeshField:
        ism: InstancedStaticMesh = in_ism.get_input()
        for name, instance_node in ism.instance_nodes.items():
            for i in range(len(instance_node.positions)):
                position = np.asarray(instance_node.positions[i])
                position += [2, 7, 15]
                instance_node.positions[i] = position.tolist()
        return InstancedStaticMeshField(ism)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup

setup(
    name='nextpcg',
    author='IEGG',
    version='0.1.2',
    license='MIT',

    description='pypapi module in NextPCG',
    long_description='''Create by AI: 
    The pypapi module in NextPCG is a powerful tool that allows users to create sophisticated programs and applications 
    with ease. It provides a comprehensive suite of features and commands that makes coding much simpler and faster. 
    With the help of this module, users can quickly and easily develop programs and applications with advanced 
    capabilities, such as natural language processing, artificial intelligence, and other complex tasks. 
    In addition, the pypapi module makes it easy to create programs and applications that are compatible with 
    different platforms and operating systems, ensuring that users can deploy their creations on a variety of platforms. 
    All in all, the pypapi module in NextPCG is a great tool for developers who want to quickly and easily create 
    complex applications.''',
    author_email='cheneyshen@tencent.com',
    url='https://glacier-request-888.notion.site/Wiki-f2600ce902b743f3ac7a40322496390a',

    # packages=setuptools.find_packages(exclude=["dson_test", "dson_generator", "dispatch"]),
    # packages=setuptools.find_packages(where=['pypapi']),
    packages=["pypapi"],
    # keywords='NLP,tokenizing,Chinese word segementation',
    # package_dir={'jieba':'jieba'},
    # package_data={'jieba':['*.*','finalseg/*','analyse/*','posseg/*']},

    install_requires=[
        "numpy >= 1.24.0",
        "Pillow >= 9.4.0",
    ],

    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        # 'License :: OSI Approved :: MIT License',
        # 'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.7',
        # 'Programming Language :: Python :: 3.8',
        # 'Programming Language :: Python :: 3.9',
        # 'Programming Language :: Python :: 3.10',

        'Topic :: Software Development :: Libraries'
    ],

    python_requires=">=3.8",
    zip_safe=True,
)


# -*- coding: utf-8 -*-
"""
Author  : NextPCG
"""

import datetime
import json
import os
import zlib

import numpy as np


class GeoBase:
    """base for mesh, heightfield"""

    def __init__(self):
        self.detail = {}
        self.vertex = {}
        self.point = {}
        self.prim = {}

    def to_json(self, io_mode=1, work_path=None):
        json_data = {}

        json_data['part_info'] = {}
        json_data['point_count'] = 1
        json_data['vertex_count'] = 1
        json_data['face_count'] = 1
        json_data['detail_count'] = 0

        json_mesh_data = {}
        json_mesh_data['point'] = self.point
        json_mesh_data['prim'] = self.prim
        json_mesh_data['vertex'] = self.vertex
        json_mesh_data['detail'] = self.detail

        json_data['attribs'] = json_mesh_data

        return json_data


"""helper functions"""


def save_output_file(output_path, json_geo_index, json_geo):
    geo_filename = json_geo_index + '_' + datetime.datetime.now().strftime("%y%m%d_%H%M%S") + '.json'
    save_path = os.path.join(output_path, geo_filename)  # .replace("\\", "/")
    save_data = {'Output_Geo': json_geo}
    with open(save_path, "w") as f:
        json.dump(save_data, f)
    return geo_filename


numpy_to_attr_data_type_remap = {}
numpy_to_attr_data_type_remap[(np.int32, 1)] = 'i32'
numpy_to_attr_data_type_remap[(np.dtype('int32'), 1)] = 'i32'
numpy_to_attr_data_type_remap[(np.float32, 1)] = 'f32'
numpy_to_attr_data_type_remap[(np.dtype('float32'), 1)] = 'f32'
# s in json
numpy_to_attr_data_type_remap[(np.float32, 2)] = 'f32x2'
numpy_to_attr_data_type_remap[(np.dtype('float32'), 2)] = 'f32x2'
numpy_to_attr_data_type_remap[(np.float32, 3)] = 'f32x3'
numpy_to_attr_data_type_remap[(np.dtype('float32'), 3)] = 'f32x3'
numpy_to_attr_data_type_remap[(np.float32, 4)] = 'f32x4'
numpy_to_attr_data_type_remap[(np.dtype('float32'), 4)] = 'f32x4'


def write_attribs_to_file(attribs_filepath, json_geo, current_offset=0):
    attribs = json_geo['attribs']
    attribs_domain_list = [attribs['vertex'], attribs['point'], attribs['prim'], attribs['detail']]
    with open(attribs_filepath, 'ab+') as f:
        for attribs_domain in attribs_domain_list:
            for attrib_name in attribs_domain:
                attr_info = attribs_domain[attrib_name]
                f.seek(current_offset)
                attr_info['offset'] = current_offset
                data = np.asarray(attr_info['data'])
                del attr_info['data']
                if data.dtype == np.float64:
                    data = data.astype(np.float32)
                if data.dtype == np.int64:
                    data = data.astype(np.int32)
                np_type = data.dtype
                shape = data.shape
                channel_count = 1
                if len(shape) == 2:
                    channel_count = int(shape[1])
                if (np_type, channel_count) in numpy_to_attr_data_type_remap:
                    data_type = numpy_to_attr_data_type_remap[(np_type, channel_count)]
                    buffer = np.asarray(data).tobytes()
                elif np.issubdtype(np_type, np.string_) or np.issubdtype(np_type, np.str_) or np.issubdtype(np_type,
                                                                                                            np.unicode_):
                    data_type = 's'
                    buffer = json.dumps(data.tolist()).encode('utf-8')  # @highyang
                else:
                    raise Exception(
                        'write_attribs_to_file error: name: {0}, type: {1}, channel_count: {2}'.format(attrib_name,
                                                                                                       np_type,
                                                                                                       channel_count))
                compress_buffer = zlib.compress(buffer)
                f.write(compress_buffer)
                attr_info['data_type'] = data_type
                attr_info['length'] = len(compress_buffer)
                attr_info['raw_length'] = len(buffer)
                current_offset += attr_info['length']
    return current_offset


attr_data_type_to_numpy_remap = {}
attr_data_type_to_numpy_remap['i32'] = (np.int32, 1)
attr_data_type_to_numpy_remap['f32'] = (np.float32, 1)
# s in json
attr_data_type_to_numpy_remap['f32x2'] = (np.float32, 2)
attr_data_type_to_numpy_remap['f32x3'] = (np.float32, 3)
attr_data_type_to_numpy_remap['f32x4'] = (np.float32, 4)


def load_attribs_from_file(attribs_filename, attr_info, work_path):
    """in binary mode, load attribs"""
    offset = attr_info['offset']
    length = attr_info['length']
    raw_length = attr_info['raw_length']
    data_type = attr_info['data_type']
    with open(os.path.join(work_path, attribs_filename), 'rb') as f:
        f.seek(offset)
        compress_buffer = f.read(length)
        raw = zlib.decompress(compress_buffer)
        if 's' in data_type:
            buffer = np.asarray(json.loads(raw))
        else:
            (np_type, channel_count) = attr_data_type_to_numpy_remap[data_type]
            buffer = np.asarray(np.frombuffer(raw, dtype=np_type))
            buffer = np.reshape(buffer, (int(len(buffer) / int(channel_count)), int(channel_count)))
            if np_type == np.int32:
                buffer = np.asarray(buffer, dtype=np.int64)  # @highyang�� change 32 -> 64
            elif np_type == np.float32:
                buffer = np.asarray(buffer, dtype=np.float64)
    return buffer.tolist()  # @highyang: use list instead np


def add_custom_attributes(geo: GeoBase, geo_json_data, skip_list, io_mode=1, work_path=""):
    if "attribs" in geo_json_data:
        input_attribs = geo_json_data["attribs"]
        input_vertex_attribs = input_attribs['vertex']
        input_point_attribs = input_attribs['point']
        input_prim_attribs = input_attribs['prim']
        input_detail_attribs = input_attribs['detail']
        input_attribs_list = [input_vertex_attribs, input_point_attribs, input_prim_attribs, input_detail_attribs]
        domain_index = -1  # used to indicate vertex, point, prim, detail
        group_index = -1

        def add_attib_data(domain_json, domain_index, input_attrib_name, input_data):
            domain_json[input_attrib_name] = {}
            domain_json[input_attrib_name]['type'] = domain_index
            domain_json[input_attrib_name]['data'] = input_data

        for input_attribs in input_attribs_list:
            domain_index += 1
            for input_attrib_name, input_attrib_val in input_attribs.items():
                if input_attrib_name in skip_list:
                    continue
                input_attrib_data = []
                if io_mode == 1:
                    input_attrib_data = input_attrib_val["data"]  # @highyang
                elif io_mode == 2 or io_mode == 3:
                    attribs_filename = geo_json_data['attribs_filename']
                    attr_info = input_attrib_val
                    input_attrib_data = load_attribs_from_file(attribs_filename, attr_info, work_path)
                if domain_index == 0:  # vertex
                    add_attib_data(geo.vertex, domain_index, input_attrib_name, input_attrib_data)
                elif domain_index == 1:  # point
                    add_attib_data(geo.point, domain_index, input_attrib_name, input_attrib_data)
                elif domain_index == 2:  # prim
                    add_attib_data(geo.prim, domain_index, input_attrib_name, input_attrib_data)
                elif domain_index == 3:  # detail
                    add_attib_data(geo.detail, domain_index, input_attrib_name, input_attrib_data)
        #     # create group
        # if group_index >= 0:
        #     group_name = 'group{0}'.format(group_index)
        #     group = geo.findPrimGroup(group_name)
        #     if group is None:
        #         group = geo.createPrimGroup(group_name)
        #     group.add(new_prims)
    return geo
